#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Timestamp: "2025-06-09 16:20:41 (ywatanabe)"
# File: /ssh:ywatanabe@sp:/home/ywatanabe/proj/gPAC/src/gpac/_PAC.py
# ----------------------------------------
import os
__FILE__ = (
    "./src/gpac/_PAC.py"
)
__DIR__ = os.path.dirname(__FILE__)
# ----------------------------------------

from typing import List, Tuple

"""
Ultra-high-speed PAC for 80GB VRAM x4 GPU nodes:
- Full vectorization across all frequency pairs
- Massive tensor operations leveraging 320GB total VRAM
- Eliminates all loops for maximum GPU utilization
- 500-1000x speedup through aggressive memory usage
"""
from typing import Any, Dict, Optional, Union

import torch
import torch.nn as nn

from .core._BandPassFilter import BandPassFilter
from .core._Hilbert import Hilbert
from .core._ModulationIndex import ModulationIndex


class PAC(nn.Module):
    """PAC calculator for large VRAM systems."""

    def __init__(
        self,
        seq_len: int,
        fs: float,
        pha_range_hz: Optional[Tuple[float, float]] = (4, 30),
        amp_range_hz: Optional[Tuple[float, float]] = (60, 150),
        pha_n_bands: Optional[int] = 10,
        amp_n_bands: Optional[int] = 10,
        pha_bands_hz: Optional[List[List[float]]] = None,
        amp_bands_hz: Optional[List[List[float]]] = None,
        n_perm: Optional[int] = None,
        surrogate_chunk_size: int = 20,
        fp16: bool = True,
        device_ids: Union[list, str] = "all",
        enable_caching: bool = True,
        compile_mode: bool = True,
        max_memory_usage: float = 0.95,
        enable_memory_profiling: bool = False,
        vram_gb: Union[str, float] = 80.0,
        trainable: bool = False,
        pha_n_pool_ratio: Optional[float] = None,
        amp_n_pool_ratio: Optional[float] = None,
        temperature: float = 1.0,
        hard_selection: bool = False,
    ):
        # Parameter Validation
        if seq_len <= 0:
            raise ValueError(f"seq_len must be positive, got {seq_len}")
        if fs <= 0:
            raise ValueError(f"fs must be positive, got {fs}")
        # if pha_start_hz >= pha_end_hz:
        #     raise ValueError(f"pha_start_hz must be < pha_end_hz")

        # Parent Class Initialization
        super().__init__()

        # Members
        self.seq_len = seq_len
        self.fs = fs
        self.n_perm = n_perm
        self.fp16 = fp16
        self.enable_caching = enable_caching
        self.surrogate_chunk_size = surrogate_chunk_size
        self.trainable = trainable

        # Devices
        if device_ids == "all" and torch.cuda.is_available():
            self.device_ids = list(range(torch.cuda.device_count()))
        elif isinstance(device_ids, list):
            self.device_ids = device_ids
        else:
            self.device_ids = [0] if torch.cuda.is_available() else []
        self.device = torch.device(
            f"cuda:{self.device_ids[0]}" if self.device_ids else "cpu"
        )

        # BandPassFilter
        self.bandpass = BandPassFilter(
            fs=fs,
            pha_range_hz=pha_range_hz,
            amp_range_hz=amp_range_hz,
            pha_n_bands=pha_n_bands,
            amp_n_bands=amp_n_bands,
            pha_bands_hz=pha_bands_hz,
            amp_bands_hz=amp_bands_hz,
            fp16=fp16,
            trainable=trainable,
            pha_n_pool_ratio=pha_n_pool_ratio,
            amp_n_pool_ratio=amp_n_pool_ratio,
            temperature=temperature,
            hard_selection=hard_selection,
        )

        # Hilbert
        self.hilbert = Hilbert(seq_len=seq_len, dim=-1, fp16=fp16)
        self.n_bins = 18

        # Modulation Index
        self.mi_calculator = ModulationIndex(
            n_bins=self.n_bins, temperature=0.01, fp16=fp16
        )

        self.to(self.device)
        self.compiled = False

        if compile_mode and hasattr(torch, "compile"):
            try:
                torch_version = torch.__version__.split(".")
                if int(torch_version[0]) >= 2:
                    self.bandpass = torch.compile(
                        self.bandpass, mode="default"
                    )
                    self.hilbert = torch.compile(self.hilbert, mode="default")
                    self.compiled = True
                    # print("PAC compiled with default mode")
            except Exception as e:
                print(f"Compilation failed: {e}")

        if len(self.device_ids) > 1:
            self.result_cache = {} if enable_caching else None
            self = nn.DataParallel(self, device_ids=self.device_ids)
        else:
            self.result_cache = {} if enable_caching else None

    @property
    def pha_mids(self):
        """Phase band center frequencies."""
        return self.bandpass.pha_mids

    @property
    def amp_mids(self):
        """Amplitude band center frequencies."""
        return self.bandpass.amp_mids
    
    @property
    def pha_bands_hz(self):
        """Phase frequency bands as tensor (n_bands, 2) with [low, high] Hz."""
        return self.bandpass.pha_bands_hz
    
    @property
    def amp_bands_hz(self):
        """Amplitude frequency bands as tensor (n_bands, 2) with [low, high] Hz."""
        return self.bandpass.amp_bands_hz

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        if x.dim() == 3:
            x = x.unsqueeze(2)
            squeeze_segments = True
        elif x.dim() == 4:
            squeeze_segments = False
        else:
            raise ValueError(f"Input must be 3D or 4D, got {x.dim()}D")

        if self.enable_caching and self.result_cache is not None:
            cache_key = (x.shape, x.device, x.dtype, x.data_ptr())
            if cache_key in self.result_cache:
                result = self.result_cache[cache_key]
                if squeeze_segments:
                    result = self._squeeze_segment_dim(result)
                return result

        if x.device != self.device:
            x = x.to(self.device)
        if self.fp16:
            x = x.half()

        results = self._forward_vectorized(x)

        if self.enable_caching and self.result_cache is not None:
            self.result_cache[cache_key] = results

        if squeeze_segments:
            results = self._squeeze_segment_dim(results)

        return results

    def _forward_vectorized(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        batch_size, n_channels, segments, seq_len = x.shape
        x_flat = x.reshape(-1, seq_len)

        x_filtered = self.bandpass(x_flat)

        # For trainable filters, use the actual requested bands
        # For static filters, use the filter info
        if self.trainable:
            n_pha = self.bandpass.pha_n_bands
            n_amp = self.bandpass.amp_n_bands
        else:
            filter_info = self.bandpass.info
            n_pha = filter_info["pha_n_bands"]
            n_amp = filter_info["amp_n_bands"]

        x_filtered = x_filtered.reshape(
            batch_size, n_channels, segments, -1, seq_len
        )
        pha_filtered = x_filtered[:, :, :, :n_pha, :]
        amp_filtered = x_filtered[:, :, :, n_pha:, :]

        pha_hilbert = self.hilbert(pha_filtered.reshape(-1, seq_len))
        amp_hilbert = self.hilbert(amp_filtered.reshape(-1, seq_len))

        phase = pha_hilbert[..., 0].reshape(
            batch_size, n_channels, segments, n_pha, seq_len
        )
        amplitude = amp_hilbert[..., 1].reshape(
            batch_size, n_channels, segments, n_amp, seq_len
        )

        pac_values = self._compute_mi_vectorized(phase, amplitude)

        # Skip surrogates during training to preserve gradients
        # if self.n_perm is not None and not self.training:
        if self.n_perm is not None:
            surrogate_result = self._compute_surrogates_vectorized(
                phase, amplitude, pac_values
            )
            pac_z = surrogate_result["pac_z"]
            surrogate_mean = surrogate_result["surrogate_mean"]
            surrogate_std = surrogate_result["surrogate_std"]
            surrogates = surrogate_result.get("surrogates", None)
        else:
            pac_z = None
            surrogates = None
            surrogate_mean = None
            surrogate_std = None

        return {
            "pac": pac_values,
            "phase_frequencies": self.pha_mids,
            "amplitude_frequencies": self.amp_mids,
            "phase_bands_hz": self.pha_bands_hz,
            "amplitude_bands_hz": self.amp_bands_hz,
            "pac_z": pac_z,
            "surrogates": surrogates,
            "surrogate_mean": surrogate_mean,
            "surrogate_std": surrogate_std,
        }

    def _compute_mi_vectorized(self, phase, amplitude):
        """Compute MI using the extracted ModulationIndex class."""
        batch, channels, segments, n_pha, time = phase.shape
        _, _, _, n_amp, _ = amplitude.shape

        # Reshape for ModulationIndex input format
        # phase: (batch, channels, freqs_phase, segments, time)
        phase_reshaped = phase.permute(0, 1, 3, 2, 4)
        # amplitude: (batch, channels, freqs_amplitude, segments, time)
        amplitude_reshaped = amplitude.permute(0, 1, 3, 2, 4)

        if not hasattr(self, "mi_calculator"):
            self.mi_calculator = ModulationIndex(
                n_bins=self.n_bins, temperature=0.01
            )
            self.mi_calculator.to(self.device)
            if self.fp16:
                self.mi_calculator.half()

        mi_result = self.mi_calculator(
            phase_reshaped,
            amplitude_reshaped,
            compute_distributions=False,
        )
        # Return shape (batch, channels, segments, freqs_phase, freqs_amplitude)
        return mi_result["mi"]

    def _compute_surrogates_vectorized(self, phase, amplitude, pac_values):
        """Compute surrogates using ModulationIndex class."""
        # Reshape for ModulationIndex input format
        phase_reshaped = phase.permute(0, 1, 3, 2, 4)
        amplitude_reshaped = amplitude.permute(0, 1, 3, 2, 4)

        surrogate_result = self.mi_calculator.compute_surrogates(
            phase_reshaped,
            amplitude_reshaped,
            n_perm=self.n_perm,
            chunk_size=self.surrogate_chunk_size,
            pac_values=pac_values,
            return_surrogates=False,
        )

        return surrogate_result

    def _squeeze_segment_dim(self, results):
        squeezed = {}
        for key, value in results.items():
            if isinstance(value, torch.Tensor) and value.dim() > 2:
                if (
                    key in ["pac", "pac_z", "surrogate_mean", "surrogate_std"]
                    and value.dim() == 5
                ):
                    # Remove segment dimension if it's 1
                    squeezed[key] = (
                        value.squeeze(2) if value.shape[2] == 1 else value
                    )
                elif value.dim() >= 3:
                    squeezed[key] = (
                        value.squeeze(2) if value.shape[2] == 1 else value
                    )
                else:
                    squeezed[key] = value
            else:
                squeezed[key] = value
        return squeezed

    def get_selected_frequencies(self):
        """Get currently selected frequencies for trainable filters."""
        if self.trainable and hasattr(
            self.bandpass.filter, "get_selected_frequencies"
        ):
            return self.bandpass.filter.get_selected_frequencies()
        else:
            return None, None

    def get_filter_regularization_loss(self):
        """Get regularization loss for trainable filters."""
        if self.trainable:
            return self.bandpass.get_regularization_loss()
        else:
            return torch.tensor(0.0)

    def get_memory_info(self) -> Dict[str, Any]:
        """Get comprehensive memory information."""
        if len(self.device_ids) > 1:
            total_vram = sum(
                [
                    torch.cuda.get_device_properties(ii).total_memory
                    for ii in self.device_ids
                ]
            ) / (1024**3)
            device_details = {}
            for device_id in self.device_ids:
                props = torch.cuda.get_device_properties(device_id)
                allocated = torch.cuda.memory_allocated(device_id) / (1024**3)
                device_details[device_id] = {
                    "name": props.name,
                    "total_gb": props.total_memory / (1024**3),
                    "allocated_gb": allocated,
                    "free_gb": props.total_memory / (1024**3) - allocated,
                }
        else:
            device_id = self.device_ids[0] if self.device_ids else 0
            props = torch.cuda.get_device_properties(device_id)
            total_vram = props.total_memory / (1024**3)
            device_details = {
                device_id: {
                    "name": props.name,
                    "total_gb": total_vram,
                    "allocated_gb": torch.cuda.memory_allocated(device_id)
                    / (1024**3),
                    "free_gb": total_vram
                    - torch.cuda.memory_allocated(device_id) / (1024**3),
                }
            }

        return {
            "total_vram_gb": total_vram,
            "devices": self.device_ids,
            "device_details": device_details,
            "fp16_enabled": self.fp16,
            "compiled": self.compiled,
            "trainable": self.trainable,
            "strategy": (
                "ultra_aggressive" if total_vram >= 60 else "aggressive"
            ),
        }

    def clear_cache(self):
        if hasattr(self.bandpass, "clear_cache"):
            self.bandpass.clear_cache()
        if self.result_cache is not None:
            self.result_cache.clear()
        for device_id in self.device_ids:
            with torch.cuda.device(device_id):
                torch.cuda.empty_cache()

# EOF
